﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

/// <summary> Read HTML file and make the white space tidy.
/// Make lines optional 72 characters.
/// </summary>
public class TidyWhite
{
	/// <summary> These are the indexes to the args[] character array
	///  and the definition of their expected contents.
	/// </summary>
	enum ARGS
	{
		FILE,
		LIMIT,
		MAX
	};
	/// <summary> default line length </summary>
	static int length_limit = 72;
	/// <summary> Save ANSI or Unicode</summary>
	static Encoding save_encoding;

	public static int Main(string[] args)
	{

		Console.Title = "TidyWhite";

		if (0 == args.Length || ARGS.MAX < (ARGS)args.Length)
		{
			HelpExit(args);
			Pause();
			return 1;
		}

		string file_arg_name = args[(int)ARGS.FILE];

		if (!File.Exists(file_arg_name))
		{
			Console.WriteLine("File does not exist: " + Path.GetFullPath(file_arg_name));
			Pause();
			return 1;
		}

		Console.WriteLine("TidyWhite: " + Path.GetFullPath(file_arg_name));

		if ((int)(ARGS.LIMIT + 1) == args.Length)
		{
			length_limit = 0;
			Int32.TryParse(args[(int)ARGS.LIMIT], out length_limit);
		}

		Console.WriteLine("Line length limit: " + length_limit);

		StringBuilder sb = new StringBuilder(Read(file_arg_name));

		sb.Replace("\r\n", "\n");
		sb.Replace("\r", "");

		// remove PRE sections before forced uppercase
		List<string> PRE_Section = Remove(sb, "PRE");

		// remove SCRIPT sections before forced uppercase
		List<string> SCRIPT_Section = Remove(sb, "SCRIPT");

		// remove STYLE sections before forced uppercase
		List<string> STYLE_Section = Remove(sb, "STYLE");

		// make sure all tags are uppercase
		int j = sb.Length;
		for (int i = 0; i < j; ++i)
		{
			if ('<' != sb[i])
				continue;
			char chr = sb[++i];
			do
			{
				chr = char.ToUpper(chr);
				sb[i] = chr;
				chr = sb[++i];
			}
			while (char.IsLetter(chr));
		}

		sb.Replace("\n", " ");
		sb.Replace("\t", " ");
		sb.Replace("    ", " ");
		sb.Replace("   ", " ");
		sb.Replace("  ", " ");

		// Newline formatting in alphabetical order by tag name
		ABipend(sb, "<!");
		SPrepend(sb, "<A");
		SAppend(sb, "</A>");
		ABipend(sb, "<BLOCKQUOTE");
		Bipend(sb, "</BLOCKQUOTE>");
		ABipend(sb, "<BODY");
		Append(sb, "<BR>");
		sb.Replace(" <BR", "<BR");
		ABipend(sb, "<DL");
		Bipend(sb, "</DL>");
		ABipend(sb, "<DT");
		ABipend(sb, "<DD");
		ABipend(sb, "<FORM");
		Bipend(sb, "</FORM>");
		ABipend(sb, "<H1");
		Bipend(sb, "</H1>");
		ABipend(sb, "<H2");
		Bipend(sb, "</H2>");
		ABipend(sb, "<H3");
		Bipend(sb, "</H3>");
		ABipend(sb, "<H4");
		Bipend(sb, "</H4>");
		ABipend(sb, "<H5");
		Bipend(sb, "</H5>");
		ABipend(sb, "<H6");
		Bipend(sb, "</H6>");
		ABipend(sb, "<HR");
		ABipend(sb, "<IMG");
		ABipend(sb, "<INPUT");
		ABipend(sb, "<LI");
		ABipend(sb, "<META");
		ABipend(sb, "<OL");
		Bipend(sb, "</OL>");
		ABipend(sb, "<P");
		ABipend(sb, "<SELECT");
		ABipend(sb, "<SCRIPT");
		// my personal special quirk for special use...
		Append(sb, "</SUP></SMALL><B>");
		ABipend(sb, "<TABLE");
		Bipend(sb, "</TABLE>");
		ABipend(sb, "<TEXTAREA");
		ABipend(sb, "<TITLE");
		Bipend(sb, "</TITLE>");
		ABipend(sb, "<TD");
		ABipend(sb, "<TH");
		ABipend(sb, "<TR");
		ABipend(sb, "<UL");
		Bipend(sb, "</UL>");

		Trim(sb);

		// we don't want any doublespace before or after above
		sb.Replace("\n\n", "\n");

		// but we do want doublespace before or after these
		Prepend(sb, "<BLOCKQUOTE");
		Append(sb, "</BLOCKQUOTE>");
		Prepend(sb, "<DL");
		Append(sb, "</DL>");
		ABipend(sb, "<FORM");
		Bipend(sb, "</FORM>");
		Prepend(sb, "<H1");
		Append(sb, "</H1>");
		Prepend(sb, "<H2");
		Append(sb, "</H2>");
		Prepend(sb, "<H3");
		Append(sb, "</H3>");
		Prepend(sb, "<H4");
		Append(sb, "</H4>");
		Prepend(sb, "<H5");
		Append(sb, "</H5>");
		Append(sb, "</H6>");
		Prepend(sb, "<HR");
		Prepend(sb, "<OL");
		Append(sb, "</OL>");
		Prepend(sb, "<P");
		ABipend(sb, "<SCRIPT");
		Prepend(sb, "<TITLE");
		Append(sb, "</TITLE>");
		Prepend(sb, "<UL");
		Append(sb, "</UL>");

		Trim(sb);

		// replace PRE sections
		Replace(sb, "PRE", PRE_Section);

		// replace SCRIPT sections
		Replace(sb, "SCRIPT", SCRIPT_Section);

		// replace STYLE sections
		Replace(sb, "STYLE", STYLE_Section);

		while ('\n' == sb[0])
			sb.Remove(0, 1);
		while ('\n' == sb[j = sb.Length - 1])
			sb.Remove(j, 1);

		WriteFile(file_arg_name, sb.ToString());

		Pause();

		return 0;
	}

	/// <summary> Build a list of removed sections
	/// Preserve a marker to return
	/// </summary>
	/// <param name="sb">string</param>
	/// <param name="tag">marker</param>
	/// <returns>list of removed sections</returns>
	static List<string> Remove(StringBuilder sb, string tag)
	{
		List<string> str = new List<string>();
		string t;
		string pre = "<" + tag;
		string post = "</" + tag;
		for (int i = 0; ;)
		{
			t = sb.ToString();
			i = t.IndexOf(pre, i, StringComparison.OrdinalIgnoreCase);
			if (0 > i)
				break;
			i += pre.Length;
			int j = t.IndexOf(post, i, StringComparison.OrdinalIgnoreCase);
			j = j - i;
			str.Add(sb.ToString(i, j));
			sb.Remove(i, j);
		}
		return str;
	}

	/// <summary> Replace removed sections
	/// </summary>
	/// <param name="sb">string</param>
	/// <param name="tag">marker</param>
	/// <returns>list of removed sections</returns>
	static void Replace(StringBuilder sb, string tag, List<string> str)
	{
		string t;
		string pre = "<" + tag;
		string post = "</" + tag;
		foreach (string tt in str)
		{
			t = sb.ToString();
			int i = t.IndexOf(pre + post);
			i += pre.Length;
			sb.Insert(i, tt);
		}
	}

	/// <summary> Standard trimming
	/// </summary>
	/// <param name="sb">string</param>
	static void Trim(StringBuilder sb)
	{
		// no space at the beginning of a line
		sb.Replace("\n  ", "\n");
		sb.Replace("\n ", "\n");

		// no space at the end of a line
		sb.Replace("  \n", "\n");
		sb.Replace(" \n", "\n");

		// maximum double space
		sb.Replace("\n\n\n\n", "\n\n");
		sb.Replace("\n\n\n", "\n\n");
	}

	static void Prepend(StringBuilder sb, string str)
	{
		sb.Replace(str, "\n" + str);
	}

	static void SPrepend(StringBuilder sb, string str)
	{
		// Protect quote, parentheses,... around anchors
		// newline after previous space if first then newline after
		string t;
		for (int i = 0; ;)
		{
			t = sb.ToString();
			i = t.IndexOf(str, i);
			if (0 > i)
				break;
			if (' ' == t[i - 1])
			{
				int j = t.IndexOf('>', i);
				sb.Insert(j + 1, "\n");
				sb.Insert(i, "\n");
			}
			else
			{
				int j = t.LastIndexOf(' ', i);
				int tj = t.LastIndexOf('>', i);
				if (j > tj)
					sb.Insert(j + 1, '\n');
			}
			i += str.Length;
		}
	}

	static void SAppend(StringBuilder sb, string str)
	{
		// Protect quote, parentheses,... around anchors
		// newline if space before avoid embarassing dangling link underline
		string t;
		for (int i = 0; ;)
		{
			t = sb.ToString();
			i = t.IndexOf(str, i);
			if (0 > i)
				break;
			int j = i + str.Length;
			if (' ' == t[i - 1])
			{
				sb.Insert(j, '\n');
				sb.Insert(i, '\n');
			}
			i = j;
		}
	}

	static void Append(StringBuilder sb, string str)
	{
		sb.Replace(str, str + "\n");
	}

	static void ABipend(StringBuilder sb, string str)
	{
		Prepend(sb, str);
		// trailing newline after the greater than (angle quote)
		string t;
		for (int i = 0; ;)
		{
			t = sb.ToString();
			i = t.IndexOf(str, i);
			if (0 > i)
				break;
			int j = t.IndexOf('>', i);
			sb.Insert(j + 1, "\n");
			i += str.Length;
		}
	}

	static void Bipend(StringBuilder sb, string str)
	{
		sb.Replace(str, "\n" + str + "\n");
	}

	/// <summary> Read the file as ANSI or UTF8 or Unicode
	/// </summary>
	/// <param name="filename">pathname of the file</param>
	/// <returns>string read</returns>
	static string Read(string filename)
	{
		// attempt UTF-8 encoding with ANSI error detection
		StreamReader sr = new StreamReader(filename, new UTF8Encoding(false, true), true);

		string str = null;
		try
		{
			str = sr.ReadToEnd();
		}
		catch
		{
		}
		if (null == str)
		{
			// use ANSI
			sr = new StreamReader(filename, Encoding.Default);
			str = sr.ReadToEnd();
		}
		save_encoding = sr.CurrentEncoding;
		sr.Close();
		return str;
	}

	/// <summary> Write the file as same original encoding
	/// </summary>
	/// <param name="file_arg_name">pathname of the file</param>
	/// <param name="str">string to write</param>
	static void WriteFile(string file_arg_name, string str)
	{
		string[] t = str.Split(new char[] { '\n' });
		StreamWriter sw = new StreamWriter(file_arg_name, false, save_encoding);
		foreach (string tt in t)
		{
			if (0 == length_limit || length_limit >= tt.Length)
			{
				sw.WriteLine(tt);
				continue;
			}
			string s = tt;
			while (length_limit < s.Length)
			{
				int j = s.LastIndexOf(' ', length_limit);
				if (0 > j)
					j = s.IndexOf(' ', length_limit);
				if (0 > j)
					break;
				sw.WriteLine(s.Substring(0, j));
				s = s.Substring(j + 1);
			}
			sw.WriteLine(s);
		}
		sw.Close();
	}

	static void HelpExit(string[] args)
	{

		Console.WriteLine("\nInvalid command line arguments.\n");
		Console.Write("TidyWhite");
		foreach (string s in args)
		{
			Console.Write(" {0}", s);
		}
		Console.WriteLine("\n\nTidy white space in HTML.\n");
		Console.WriteLine("Usage:\n");
		Console.WriteLine("\tTidyWhite filename [limit_option]\n");
		Console.WriteLine("\tfilename to process\n");
		Console.WriteLine("\tlimit_option to change 72 character line limit");
		Console.WriteLine("\t(0 or non-numeric means to bypass line limit)\n");
		Console.WriteLine("Example:\n");
		Console.WriteLine("\tTidyWhite TidyWhite.html");
	}

	static void Beep()
	{
		// A pleasing melodic beep
		Console.Beep(880, 10);
		Console.Beep(440, 10);
	}

	static void Pause()
	{
		// for console run without command line when displaying an error message
		Beep();
		Console.WriteLine();
		Console.Write("Type the ENTER key when ready...");
		Console.ReadLine();
		Console.WriteLine();
	}

}
